/*****************************************************************************
* $Id: nf2_download.c 4193 2008-06-23 19:26:29Z grg $
*
* Project: NetFPGA2, Stanford University
* This program reads a Xilinx .bin or .bit file and downloads it to a CNET 
* or CPCI device on a NF2 board. (CNET is a Virtex 2 Pro. CPCI is a Spartan 2)
*
* Usage: ./nf2_download <code filename> 
*
* Changes: Modified for Windows Application
*
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

#include "nf2_download.h"
#include "..\..\inc\nf2.h"
#include "util.h"
#include "..\..\inc\public.h"

#include "..\..\inc\reg_defines.h"
#include "getopt.h"

#define DEFAULT_IFACE	"nf2c0"

/* Global vars */
HANDLE hDevice;
//char * bitfilename;
//static struct nf2device nf2;
void download ();

/*
Checks for commandline args, intializes globals, then  begins code download.
*/
int main(int argc, char **argv) {

	processArgs(argc, argv);

	hDevice = RequestHandler ();
	if ( !hDevice) {
		printf ("failed to open device");
		return 0;
	}

	download();

	CloseHandle (hDevice);
	return SUCCESS;
}

void download () {

	PCI_COMMON_CONFIG config;
	PPCI_COMMON_CONFIG pConfig = &config;

	// if -c we should preserve the configuration data first
	if (cpci_reprog) {
		if ( !GetPCIInfo ( hDevice, pConfig))
			exit(1);
		printf ("vendor ID is: %x\n", pConfig->VendorID);
		printf ("device ID is: %x\n", pConfig->DeviceID);
		printf ("Header type is: %x\n", pConfig->HeaderType);
		printf ("Base address 0 is: %x\n", pConfig->u.type0.BaseAddresses[0]);
	}

	InitGlobals();

	BeginCodeDownload(bin_file_name);

	if (!cpci_reprog)
		ResetDevice();

	// now we should restore the configuration data
	if (cpci_reprog) {
		Sleep (1500);
		if ( !SetPCIInfo ( hDevice, pConfig))
			exit(1);
	}
	// read the configuration data again, it should be the same as the first time
	if (cpci_reprog) {
		if ( !GetPCIInfo ( hDevice, pConfig))
			exit(1);
		printf ("vendor ID is: %x\n", pConfig->VendorID);
		printf ("device ID is: %x\n", pConfig->DeviceID);
		printf ("Header type is: %x\n", pConfig->HeaderType);
		printf ("Base address 0 is: %x\n", pConfig->u.type0.BaseAddresses[0]);
	}

	/* reset the PHYs */
	NF2_WR32( MDIO_0_CONTROL_REG, 0x8000);
	Sleep (100);
	NF2_WR32( MDIO_1_CONTROL_REG, 0x8000);
	Sleep (100);
	NF2_WR32( MDIO_2_CONTROL_REG, 0x8000);
	Sleep (100);
	NF2_WR32( MDIO_3_CONTROL_REG, 0x8000);
	Sleep (100);

	VerifyDevInfo();
}

/*
Initializes the global variables this code uses.
*/
void InitGlobals() {
	bytes_sent = 0;
}


/*
Starts the code download process.  First opens the file containing the code to
download.  Loops to do a single iteration on the state machine.  state is tracked
via the global variable "serverState".  This loop continues until we error and exit,
or the end state is reached.
*/
void BeginCodeDownload(char *codefile_name) {

	FILE *codefile; 
	if ((codefile = fopen(codefile_name, "rb")) == NULL) {
		printf( "Failed to open %s.\n", codefile_name);
		FatalError(); //exit
	}

	StripHeader(codefile);
	DownloadCode(codefile);

	fclose(codefile);
}


/*
* Strip the header off a bin file
*
* Does so by reading past the header
*/
void StripHeader(FILE *code_file) {
	size_t code_header_size;
	unsigned char * code_header;

	int header_len;

	/* Malloc memory for the header */
	code_header = (unsigned char *) malloc(sizeof(unsigned char) * READ_BUFFER_SIZE);

	/* Read the first few bytes */
	code_header_size = fread(code_header, sizeof(unsigned char), 2, code_file);

	/* Check to see if we're dealing with a header or not */
	if (code_header_size && (code_header[0] != 0xff || code_header[1] != 0xff)) {
		header_len = code_header[0] << 8 | code_header[1];

		/* Read the header and skip a field */
		code_header_size = fread(code_header, sizeof(unsigned char), header_len + 2 + 1 + 2, code_file);

		/* Read the ncd file name */
		header_len = code_header[header_len + 2 + 1] << 8 | code_header[header_len + 2 + 1 + 1];
		code_header_size = fread(code_header, sizeof(unsigned char), header_len + 1 + 2, code_file);
		printf("Bit file built from: %s\n", code_header);

		/* Read the part name */
		header_len = code_header[header_len + 1] << 8 | code_header[header_len + 1 + 1];
		code_header_size = fread(code_header, sizeof(unsigned char), header_len + 1 + 2, code_file);
		printf("Part: %s\n", code_header);

		/* Read the date */
		header_len = code_header[header_len + 1] << 8 | code_header[header_len + 1 + 1];
		code_header_size = fread(code_header, sizeof(unsigned char), header_len + 1 + 2, code_file);
		printf("Date: %s\n", code_header);

		/* Read the time */
		header_len = code_header[header_len + 1] << 8 | code_header[header_len + 1 + 1];
		code_header_size = fread(code_header, sizeof(unsigned char), header_len + 1 + 4, code_file);
		printf("Time: %s\n", code_header);
	}
	else {
		rewind(code_file);
	}

	/* Free the code_header variable */
	free(code_header);
}

/*
Download the codefile by writing it to the programming register in CPCI.
*/
void DownloadCode(FILE *code_file) {

	ULONG result;
	ULONG version;
	size_t code_data_size;
	unsigned char * code_data;
	ULONG retries;
	int bytes_expected;
	int count=0;

	/*
	* Identify what version of the board we are running
	*/
	version = NF2_RD32(CPCI_ID);
	version &= 0xffffff;


	if (!cpci_reprog) {
		/* 
		First, make sure the download interface is reset.
		This flushes buffers, resets state machines etc.
		*/

		NF2_WR32 ( CPCI_PROGRAMMING_CONTROL, 1);
		result = NF2_RD32( CPCI_PROGRAMMING_CONTROL);
		printf ( "control is now: x%x\n", result);

		/* Clear the error registers */
		NF2_WR32 ( CPCI_ERROR, 0);

		/* Wait a while for the PROG_B cycle to finish. */
		Sleep(100);

		/* 
		Read programming status:
		Check that DONE bit (bit 8) is zero and FIFO (bit 1) is empty (1).
		*/

		result = NF2_RD32( CPCI_PROGRAMMING_CONTROL);
		printf ( "control is now: x%x\n", result);

		result = NF2_RD32( CPCI_PROGRAMMING_STATUS);

		if ((result & 0x102) != 0x2) {
			printf( "After resetting Programming interface, expected status to be 1 (FIFO empty).\n");
			printf( "However status & 0x102 is 0x%0x\n", (result & 0x102));
			FatalError();
		}

		/* Check the error register */
		printf( "Error Registers: %x\n", NF2_RD32(CPCI_ERROR));

		printf( "Good, after resetting programming interface the FIFO is empty\n");

		/* Sleep for a while to allow the INIT pin to be reset */
		Sleep(10000);
		retries = 3;

		while ( NF2_RD32(CPCI_PROGRAMMING_STATUS) & 0x10002)
		{
			if ( NF2_RD32(CPCI_PROGRAMMING_STATUS) & 0x10000)
				Sleep(10000);
			else
				break;
			retries--;

			if (retries <= 0)
			{
				printf("CPCI's INIT signal did not clear in time, exiting\n");
				printf ("CPCI_PROGRAMMING_STATUS: 0x%08x\n", NF2_RD32(CPCI_PROGRAMMING_STATUS));
				exit(1);
			}      
		}
	} // if (!cpci_reprog)

	/*
	Read the code file, and write it to the card.
	*/
	printf ("Read the code file, and write it to the card.\n");
	code_data = (unsigned char *) malloc(sizeof(unsigned char) * READ_BUFFER_SIZE);

	/* check num bytes read */
	while (code_data_size = fread(code_data, sizeof(unsigned char), READ_BUFFER_SIZE, code_file)) 
	{
		if (cpci_reprog)
			DownloadCPCICodeBlock(code_data, code_data_size);
		else
			DownloadVirtexCodeBlock(code_data, code_data_size);
		bytes_sent += code_data_size;
	}

	/* Free the code_data variable */
	free(code_data);

	/* Work out how large the download should have been */
	if (cpci_reprog) {
		bytes_expected = CPCI_BIN_SIZE;
	}
	else {
		switch (version) {
		 case 1: bytes_expected = VIRTEX_BIN_SIZE_V2_0; break;
		 case 2: bytes_expected = VIRTEX_BIN_SIZE_V2_1; break;
		 case 3: bytes_expected = VIRTEX_BIN_SIZE_V2_1; break;
		 default : bytes_expected = -1;
		}
	}

	printf( "Download completed -  %d bytes. (expected %d).\n", bytes_sent, bytes_expected);

	/*
	* Kick off the programming process or wait for the process to complete
	*/
	if (cpci_reprog) {
		NF2_WR32( VIRTEX_PROGRAM_CTRL_ADDR, DISABLE_RESET | START_PROGRAMMING);
		printf( "Instructed CPCI reprogramming to start. Please reload PCI BARs.\n");
	}
	else {
		/*
		Now wait to see that DONE goes high (bit 8) and INIT (16) is low
		*/
		for (retries=0; retries<3; retries++) {
			Sleep(1);
			result = NF2_RD32( CPCI_PROGRAMMING_STATUS);
			if  ((result & 0x100) == 0x100 ) { 
				printf( "DONE went high - chip has been successfully programmed.\n");
				return;
			}
			if  ((result & 0x10000) == 0x10000 ) { 
				printf( "INIT went high - appears to be a programming error.\n");
				FatalError();
			}
			if (retries == 2) {
				printf( "DONE has not gone high - looks like an error\n");
				FatalError();
			}
		}
	}
}


/*
Given a block of bytes to write, and a bytecount, create 32 bit words from these
bytes and write them to the Programming FIFO.
After each 8 words stop and check that the FIFO is empty.
*/
void DownloadVirtexCodeBlock (u_char *code_data, size_t code_data_size) {

	ULONG result;
	ULONG data_word;
	size_t bytes_left;
	ULONG count = 0;

	bytes_left = code_data_size;

	while (bytes_left)
	{
		data_word = (ULONG) (*code_data++); bytes_left--;
		if (bytes_left) { data_word |= ((ULONG) (*code_data++))<<8 ; bytes_left--;}
		if (bytes_left) { data_word |= ((ULONG) (*code_data++))<<16 ; bytes_left--;}
		if (bytes_left) { data_word |= ((ULONG) (*code_data++))<<24 ; bytes_left--;}

		NF2_WR32( CPCI_PROGRAMMING_DATA, data_word);

		/* Every 8 words we need to check the FIFO - should always be empty!
		* (or the done flag should be asserted)
		*/
		if (++count == 8) {
			count = 0;
	
			while (((result=NF2_RD32(CPCI_PROGRAMMING_STATUS)) & 0x10002) != 2 && (result & 0x100) != 0x100) {
				if (result & 0x10000) {
					printf( "INIT went active during programming - there was an error!\n");
					exit(1);
				}
				printf( "Strange. FIFO wasnt empty... trying again.\n");
				Sleep(100);
				result = NF2_RD32 ( CPCI_PROGRAMMING_STATUS);

				if ((result & 0x10002) != 2) {
					printf( "Retrying ... FIFO still not empty. Giving up.\n");
					printf( "Last status word read was 0x%0x\n", result);
				}
			}
		}
	}
}


/*
Given a block of bytes to write, and a bytecount, create 32 bit words from these
bytes and write them to the programming RAM in the Virtex
*/
void DownloadCPCICodeBlock (u_char *code_data, size_t code_data_size) {

	ULONG data_word;
	size_t bytes_left;
	ULONG count = 0;

	bytes_left = code_data_size;

	while (bytes_left)
	{
		data_word = ((ULONG) (*code_data++)) << 24; bytes_left--;
		if (bytes_left) { data_word |= ((ULONG) (*code_data++))<<16 ; bytes_left--;}
		if (bytes_left) { data_word |= ((ULONG) (*code_data++))<<8  ; bytes_left--;}
		if (bytes_left) { data_word |= ((ULONG) (*code_data++))     ; bytes_left--;}

		NF2_WR32( prog_addr, data_word);
		prog_addr += 4;
	}
}

/*
* Reset the device
*/
void ResetDevice(void) {
	ULONG val;

	/* Read the current value of the control register so that we can modify
	* it to do a reset */
	val = NF2_RD32(CPCI_CTRL);

	/* Write to the control register to reset it */
	NF2_WR32(CPCI_CTRL, val | 0x100);

	/* Sleep for a while to let the reset complete */
	Sleep(2);
}



/*
* Verify the device info
*/
void VerifyDevInfo(void) {

	if (!ignore_dev_info && !cpci_reprog) {
		/* Read the device info from the board */
		NF2ReadInfo ( hDevice);

		/* Check the CPCI version info */
		if (nf2_device_id == -1) {
			printf( "WARNING: NF2 device info not found. Cannot verify that the CPCI version matches.\n");
		}
		else {
			printf( "Downloaded design:\n");
			printf( "  %s\n", nf2_device_str);
			printf( "  Device ID: %d\n", nf2_device_id);
			printf( "  Revision : %d\n", nf2_revision);

			if (cpci_version != nf2_cpci_version ||
				cpci_revision != nf2_cpci_revision) {
					printf( "Error: Virtex design compiled against a different CPCI version\n");
					printf( "  Active CPCI version: %d rev %d\n", cpci_version, cpci_revision);
					printf( "  NF2 built against  : %d rev %d\n", nf2_cpci_version, nf2_cpci_revision);
					exit(1);
			}
			else {
				printf("Virtex design compiled against active CPCI version\n");
			}
		}
	}
}



/*
Called once we have encountered a fatal error.
Leave message in log file and terminate.
*/          
void FatalError() {
	printf( "\nFatal Error, exiting...\n");
	exit(FAILURE);
}

/*
* readReg - read a register
*/
ULONG NF2_RD32(ULONG addr)
{
	ULONG val;

	if (!ReadRegister(hDevice, addr, &val))
	{
		printf( "Error reading register %x\n", addr);
		exit(1);
	}

	return val;
}

void NF2_WR32(ULONG addr, ULONG data)
{
	if (!WriteRegister(hDevice, addr, &data))
	{
		printf( "Error writing register %x\n", addr);
		exit(1);
	}
}

/* 
Process the arguments.
*/

void processArgs (int argc, char **argv ) {

	char c;

	/* set defaults */
	verbose = 0;
	cpci_reprog = 0;
	ignore_dev_info = 0;
	prog_addr = VIRTEX_PROGRAM_RAM_BASE_ADDR;

	/* don't want getopt to moan - I can do that just fine thanks! */
	opterr = 0;

	while ((c = getopt (argc, argv, "vcn:i:")) != -1)
		switch (c)
	{
		case 'v':
			verbose = 1;
			break;
		case 'c':
			cpci_reprog = 1;
			break;
		case 'n':
			ignore_dev_info = 1;
			break;
		case '?':
			printf ("Unknown option character.\n");
		default:
			usage();
			exit(1);
	}



	/* optind starts at first non-option argument (filename of bin file) */

	if (argv[optind] == NULL) {
		usage(); exit(1);
	}

	/* filename MUST end in .bin or .bit */
	if (strstr(argv[optind],".bin") == NULL && strstr(argv[optind],".bit") == NULL ) {
		fprintf(stderr,"Error: the filename must end in .bin or .bit\n");
		exit(1);
	}

	bin_file_name = argv[optind];

	if (verbose) {
		printf ("logfile = %s.   bin file = %s\n", log_file_name, bin_file_name);
	}

}


/*
Describe usage of this program.
*/
void usage () {
	printf("Usage: ./nf2_download <options>  [filename.bin | filename.bit]\n");
	printf("\nOptions: -l <logfile> (default is stdout).\n");
	printf("         -i <iface> : interface name.\n");
	printf("         -c : reprogram CPCI.\n");
	printf("         -n : don't verify Spartan/Virtex build compatibility.\n");
	printf("         -v : be verbose.\n");
}


/* vim:set shiftwidth=3 softtabstop=3 expandtab: */
