/*++

Copyright (c) 2009	Microsoft Corporation

Module Name:
	Device.c

Abstract:
	PCI Driver Routine Functions

Revision History:
	Who 		When		What						
	--------	--------	-----------------------------------------------
	Chen Tian	16-03-09	Wrtitten For NetFPGA2 Card in Microsoft Research Asia
Notes:

--*/
#include "..\inc\precomp.h"
#include "..\inc\Device.h"
#include "..\inc\DeviceIo.h"
#include "..\inc\IsrDpc.h"
#include "..\inc\dbg.h"
#include "..\inc\reg_defines.h"
#include "..\inc\nf2.h"
#include "..\inc\nf2kernel.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text (PAGE, PCIEEvtDeviceAdd)
#pragma alloc_text (PAGE, PCIEEvtDevicePrepareHardware)
#pragma alloc_text (PAGE, PCIEEvtDeviceReleaseHardware)
#pragma alloc_text (PAGE, PCIEEvtDeviceD0Entry)
#pragma alloc_text (PAGE, PCIEEvtDeviceD0Exit)
#pragma alloc_text (PAGE, PCIESetIdleAndWakeSettings)
#endif

DEFINE_GUID(GUID_PCIE_INTERFACE, 
	0x133b9c58, 0x7988, 0x4bf2, 0xb2, 0x34, 0xfa, 0xad, 0x8a, 0x74, 0xc8, 0xfb);


NTSTATUS
PCIEEvtDeviceAdd(
	IN WDFDRIVER		Driver,
	IN PWDFDEVICE_INIT	DeviceInit
	)
/*++

Routine Description:

EvtDriverDeviceAdd is called by the framework in response to AddDevice
call from the PnP manager. Here the driver should register all the
PNP, power and Io callbacks, register interfaces and allocate other
software resources required by the Device. The driver can query
any interfaces or get the config space information from the bus driver
but cannot access hardware registers or initialize the Device.

Arguments:

Return Value:

--*/
{
	NTSTATUS						Status = STATUS_SUCCESS;
	WDF_PNPPOWER_EVENT_CALLBACKS	PnpPowerCallbacks;
	WDF_OBJECT_ATTRIBUTES			Attributes;
	WDFDEVICE						Device;
	PDEVICE_EXTENSION				pDevExt = NULL;
	UNICODE_STRING					DeviceName;
	UNICODE_STRING					DeviceSymbolLink;
	ULONG							isUpperEdgeNdis;

	DBGPRINT(MP_TRACE, ("---> PCIEEvtDeviceAdd"));

	UNREFERENCED_PARAMETER(Driver);

	PAGED_CODE();

	WdfDeviceInitSetIoType(DeviceInit, WdfDeviceIoDirect);

	//
	// Find out whether the driver is installed as a miniport.
	//
	PciDrvReadRegistryKeyValue(DeviceInit,
								  L"IsUpperEdgeNdis",
								  &isUpperEdgeNdis
								  );
	
	DBGPRINT(MP_INFO, ("Upper edge is %sNDIS\n", isUpperEdgeNdis?"":"not "));

	//
	// Zero out the PnpPowerCallbacks structure.
	//
	WDF_PNPPOWER_EVENT_CALLBACKS_INIT(&PnpPowerCallbacks);

	//
	// Set Callbacks for any of the functions we are interested in.
	// If no callback is set, Framework will take the default action
	// by itself.
	//
	PnpPowerCallbacks.EvtDevicePrepareHardware = PCIEEvtDevicePrepareHardware;
	PnpPowerCallbacks.EvtDeviceReleaseHardware = PCIEEvtDeviceReleaseHardware;

	//
	// These two callbacks set up and tear down hardware state that must be
	// done every time the Device moves in and out of the D0-working state.
	//
	PnpPowerCallbacks.EvtDeviceD0Entry			  = PCIEEvtDeviceD0Entry;
	PnpPowerCallbacks.EvtDeviceD0Exit			  = PCIEEvtDeviceD0Exit;

	//
	// Register the PnP Callbacks..
	//
	WdfDeviceInitSetPnpPowerEventCallbacks(DeviceInit, &PnpPowerCallbacks);

	if(isUpperEdgeNdis) 
	{
		//
		// If upper edge is NDIS then relinquish the power policy ownership
		// because NDIS always assumes that it's default policy power owner.
		//
		WdfDeviceInitSetPowerPolicyOwnership(DeviceInit, FALSE);
	} 
	else 
	{
		// Since we are the function driver, we are now the power policy owner
		// for the device according to the default framework rule. We will register
		// our power policy callbacks after finding the wakeup capability of the device.
	}

	//
	// Initialize Fdo Attributes.
	//
	WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&Attributes, DEVICE_EXTENSION);
	//
	// By opting for SynchronizationScopeDevice, we tell the framework to
	// synchronize callbacks events of all the objects directly associated
	// with the Device. In this driver, we will associate queues and
	// and DpcForIsr. By doing that we don't have to worry about synchronizing
	// access to Device-context by Io Events and DpcForIsr because they would
	// not concurrently ever. Framework will serialize them by using an
	// internal Device-lock.
	//
	Attributes.SynchronizationScope = WdfSynchronizationScopeDevice;

	RtlInitUnicodeString(&DeviceName, BUDDY_DEVICE_NAME);
	RtlInitUnicodeString(&DeviceSymbolLink, SDR_DEVICE_SYMBOL_NAME);

	Status = WdfDeviceInitAssignName(
		DeviceInit,
		&DeviceName
		);

	if (!NT_SUCCESS(Status)) 
	{
		return Status;
	}
	
	//
	// Create the Device
	//
	Status = WdfDeviceCreate( &DeviceInit, &Attributes, &Device );

	if (!NT_SUCCESS(Status)) 
	{
		return Status;
	}

	Status = WdfDeviceCreateSymbolicLink(Device, &DeviceSymbolLink);
	
	if (!NT_SUCCESS(Status)) 
	{
		return Status;
	}

	//
	// Get the DeviceExtension and initialize it. PCIEGetDeviceContext is an inline function
	// defined by WDF_DECLARE_CONTEXT_TYPE_WITH_NAME macro in the
	// private header file. This function will do the type checking and return
	// the Device context. If you pass a wrong object a wrong object handle
	// it will return NULL and assert if run under framework verifier mode.
	//
	pDevExt = PCIEGetDeviceContext(Device);

	pDevExt->Device = Device;

	//
	// Device creation is complete.
	// Get the DeviceExtension and initialize it.
	//
	pDevExt->IsUpperEdgeNdis = isUpperEdgeNdis ? TRUE: FALSE;

   	DBGPRINT(MP_INFO, (
				"PDO(0x%p) FDO(0x%p), Lower(0x%p)\n",
				WdfDeviceWdmGetPhysicalDevice (pDevExt->Device),
				WdfDeviceWdmGetDeviceObject (pDevExt->Device),
				WdfDeviceWdmGetAttachedDevice(pDevExt->Device)));


	//
	// Tell the Framework that this Device will need an interface
	// NOTE: See the note in Public.h concerning this GUID value.
	//
	Status = WdfDeviceCreateDeviceInterface( Device, (LPGUID) &GUID_PCIE_INTERFACE, NULL);

	if (!NT_SUCCESS(Status)) 
	{
		return Status;
	}

	//
	// Set the idle and wait-wake policy for this Device.
	//
	// @@ It seems our FPGA does not support this
	// PCIESetIdleAndWakeSettings(pDevExt);

	//
	// Initalize the Device Extension.
	//
	Status = PCIEInitializeDeviceExtension(pDevExt);

	if (!NT_SUCCESS(Status)) 
	{		  
		return Status;
	}

	DBGPRINT(MP_TRACE, ("<--- PCIEEvtDeviceAdd"));

	return Status;
}

NTSTATUS
PCIEInitializeDeviceExtension(IN PDEVICE_EXTENSION pDevExt)
/*++
Routine Description:

This routine is called by EvtDeviceAdd. Here the device context is
initialized and all the software resources required by the device is
allocated.

Arguments:

pDevExt 	Pointer to the Device Extension

Return Value:

NTSTATUS

--*/
{
	NTSTATUS				Status;
	WDF_IO_QUEUE_CONFIG  	QueueConfig;
	WDF_IO_QUEUE_CONFIG  	InternalQueueConfig;

	DBGPRINT(MP_INFO, ("---> PCIEInitializeDeviceExtension"));

	PAGED_CODE();

	//
	// Setup a queue to handle only IRP_MJ_WRITE requests in Sequential
	// dispatch mode. This mode ensures there is only one write request
	// outstanding in the driver at any time. Framework will present the next
	// request only if the current request is completed.
	// Since we have configured the queue to dispatch all the specific requests
	// we care about, we don't need a default queue.  A default queue is
	// used to receive requests that are not preconfigured to goto
	// a specific queue.
	//
	WDF_IO_QUEUE_CONFIG_INIT ( &QueueConfig, WdfIoQueueDispatchSequential);

	QueueConfig.EvtIoDeviceControl = PCIEDeviceIoCtrl;

	Status = WdfIoQueueCreate(
				pDevExt->Device,
				&QueueConfig,
				WDF_NO_OBJECT_ATTRIBUTES,
				&pDevExt->IOCtrlQueue 
				);
	
	if(!NT_SUCCESS(Status)) 
	{
		return Status;
	}

	// causes the framework to queue a specified type of I/O requests to a specified I/O queue
	Status = WdfDeviceConfigureRequestDispatching(
				pDevExt->Device,
				pDevExt->IOCtrlQueue,
				WdfRequestTypeDeviceControl
				);
	
	if(!NT_SUCCESS(Status)) 
	{
		return Status;
	}

	//
	// Manual internal queue for device I/O control requests. This will be used to
	// queue the ioctl requests presented to us from the parallel ioctl queue
	// when we cannot handle them immediately. This is a managed queue.
	// Requests be get canceled while waiting in the queue without any
	// notification to the driver.
	//

	WDF_IO_QUEUE_CONFIG_INIT(
		&QueueConfig,
		WdfIoQueueDispatchManual
		);

	Status = WdfIoQueueCreate (
				   pDevExt->Device,
				   &QueueConfig,
				   WDF_NO_OBJECT_ATTRIBUTES,
				   &pDevExt->PendingIoctlQueue
				   );

	if(!NT_SUCCESS (Status))
	{
		DBGPRINT(MP_INFO, ( "Error Creating manual Ioctl Queue 0x%x\n", Status));
		return Status;
	}

	//
	// We will create and configure a queue for receiving
	// write requests. If these requests have to be pended for any
	// reason, they will be forwarded to a manual queue created after this one.
	// Framework automatically takes the responsibility of handling
	// cancellation when the requests are waiting in the queue. This is
	// a managed queue. So the framework will take care of queueing
	// incoming requests when the pnp/power state transition takes place.
	// Since we have configured the queue to dispatch all the specific requests
	// we care about, we don't need a default queue. A default queue is
	// used to receive requests that are not preconfigured to go to
	// a specific queue.
	//
	WDF_IO_QUEUE_CONFIG_INIT(
		&QueueConfig,
		WdfIoQueueDispatchSequential
		);

	QueueConfig.EvtIoWrite = PciDrvEvtIoWrite;

	Status = WdfIoQueueCreate(
				pDevExt->Device,
				&QueueConfig,
				WDF_NO_OBJECT_ATTRIBUTES,
				&pDevExt->WriteQueue // queue handle
			 	);

	if (!NT_SUCCESS (Status)) 
	{
		DBGPRINT(MP_ERROR, ("WdfIoQueueCreate failed 0x%x\n", Status));
		return Status;
	}

	 Status = WdfDeviceConfigureRequestDispatching(
					pDevExt->Device,
					pDevExt->WriteQueue,
					WdfRequestTypeWrite);

	if(!NT_SUCCESS (Status))
	{
		ASSERT(NT_SUCCESS(Status));
		DBGPRINT(MP_ERROR, ("Error in config write Queue 0x%x\n", Status));
		return Status;
	}

	//
	// Manual internal queue for write reqeusts. This will be used to queue
	// the write requests presented to us from the parallel default queue
	// when we are low in TCB resources or when the device
	// is busy doing link detection.
	// Requests can be canceled while waiting in the queue without any
	// notification to the driver.
	//
	WDF_IO_QUEUE_CONFIG_INIT(
		&QueueConfig,
		WdfIoQueueDispatchManual
		);

	Status = WdfIoQueueCreate (
				   pDevExt->Device,
				   &QueueConfig,
				   WDF_NO_OBJECT_ATTRIBUTES,
				   &pDevExt->PendingWriteQueue
				   );

	if(!NT_SUCCESS (Status)){
		DBGPRINT(MP_ERROR, ("Error Creating manual write Queue 0x%x\n", Status));
		return Status;
	}

	//
	// Manual internal queue for read reqeusts. This will be used to queue
	// the read requests presented to us from the parallel default queue
	// when we are low in TCB resources or when the device
	// is busy doing link detection.
	// Requests can be canceled while waiting in the queue without any
	// notification to the driver.
	//
	WDF_IO_QUEUE_CONFIG_INIT(
		&QueueConfig,
		WdfIoQueueDispatchManual
		);

	Status = WdfIoQueueCreate (
				   pDevExt->Device,
				   &QueueConfig,
				   WDF_NO_OBJECT_ATTRIBUTES,
				   &pDevExt->PendingReadQueue
				   );

	if(!NT_SUCCESS (Status))
	{
		DBGPRINT(MP_ERROR, ("Error Creating manual read Queue 0x%x\n", Status));
		return Status;
	}

	Status = WdfDeviceConfigureRequestDispatching(
					pDevExt->Device,
					pDevExt->PendingReadQueue,
					WdfRequestTypeRead
					);

	if(!NT_SUCCESS (Status))
	{
		ASSERT(NT_SUCCESS(Status));
		DBGPRINT(MP_ERROR, ("Error in config read Queue 0x%x\n", Status));
		return Status;
	}

	//
	// Create a WDFINTERRUPT object.
	//
	Status = PCIEInterruptCreate(pDevExt);

	if (!NT_SUCCESS(Status)) 
	{
		return Status;
	}

	DBGPRINT(MP_INFO, ("<--- PCIEInitializeDeviceExtension"));

	return Status;
}

/*++
Routine Description:

Called by EvtDeviceAdd to set the idle and wait-wake policy. Registering this policy
causes Power Management Tab to show up in the Device manager. By default these
options are enabled and the user is provided control to change the settings.

Return Value:

NTSTATUS - Failure Status is returned if the Device is not capable of suspending
or wait-waking the machine by an external event. Framework checks the
capability information reported by the bus driver to decide whether the Device is
capable of waking the machine.

--*/
NTSTATUS
PCIESetIdleAndWakeSettings(IN PDEVICE_EXTENSION FdoData)
{
	WDF_DEVICE_POWER_POLICY_IDLE_SETTINGS		IdleSettings;
	WDF_DEVICE_POWER_POLICY_WAKE_SETTINGS	WakeSettings;
	NTSTATUS	Status = STATUS_SUCCESS;

	PAGED_CODE();

	//
	// Init the idle policy structure.
	//
	WDF_DEVICE_POWER_POLICY_IDLE_SETTINGS_INIT(&IdleSettings, IdleCanWakeFromS0);
	IdleSettings.IdleTimeout = 10000; // 10-sec

	Status = WdfDeviceAssignS0IdleSettings(FdoData->Device, &IdleSettings);

	if ( !NT_SUCCESS(Status)) 
	{
		return Status;
	}

	//
	// Init wait-wake policy structure.
	//
	WDF_DEVICE_POWER_POLICY_WAKE_SETTINGS_INIT(&WakeSettings);

	Status = WdfDeviceAssignSxWakeSettings(FdoData->Device, &WakeSettings);

	if (!NT_SUCCESS(Status)) 
	{
		return Status;
	}

	return Status;
}

NTSTATUS
PCIEEvtDevicePrepareHardware (
	WDFDEVICE			Device,
	WDFCMRESLIST		Resources,
	WDFCMRESLIST		ResourcesTranslated
	)
/*++

Routine Description:

Performs whatever initialization is needed to setup the Device, setting up
a DMA channel or mapping any I/O port resources.  This will only be called
as a Device starts or restarts, not every time the Device moves into the D0
state.	Consequently, most hardware initialization belongs elsewhere.

Arguments:

Device - A handle to the WDFDEVICE

Resources - The raw PnP resources associated with the Device.  Most of the
time, these aren't useful for a PCI Device.

ResourcesTranslated - The translated PnP resources associated with the
Device.  This is what is important to a PCI Device.

Return Value:

NT Status code - failure will result in the Device stack being torn down

--*/
{
	NTSTATUS				Status = STATUS_SUCCESS;
	HRESULT 				hRes   = 1;
	PDEVICE_EXTENSION		pDevExt;
	ULONG					Index;
	BOOLEAN 				FoundAddress   = FALSE;
	ULONG					RegPhyAddressLength  = 0;
	PHYSICAL_ADDRESS		RegPhyAddressBase	 = {0, 0};
	PHYSICAL_ADDRESS		PhyLow = {0,0};
	PHYSICAL_ADDRESS		PhyHigh = {0x80000000, 0};

	WDF_DMA_ENABLER_CONFIG			EnablerConfig;
	PCM_PARTIAL_RESOURCE_DESCRIPTOR	pResDesc;
	WDF_OBJECT_ATTRIBUTES				attributes;

	UNREFERENCED_PARAMETER(Resources);

	DBGPRINT(MP_TRACE, ("---> PCIEEvtDevicePrepareHardware"));

	PAGED_CODE();	 

	do 
	{
		pDevExt = PCIEGetDeviceContext(Device);
		//
		// Parse the resource list and save the resource information.
		//
		for (Index=0; Index < WdfCmResourceListGetCount(ResourcesTranslated); Index++) 
		{
			pResDesc = WdfCmResourceListGetDescriptor(ResourcesTranslated, Index);

			if(!pResDesc) 
			{
				return STATUS_DEVICE_CONFIGURATION_ERROR;
			}

			switch (pResDesc->Type) 
			{
			case CmResourceTypeMemory:
				if (!FoundAddress )
				{
					RegPhyAddressBase = pResDesc->u.Memory.Start;
					RegPhyAddressLength = pResDesc->u.Memory.Length;
					FoundAddress	= TRUE;
				}
				break;

		  	// Do we need to handle other resources here?

			default:
				//
				// Ignore all other descriptors
				//
				break;
			}
		}

		if (!FoundAddress)
		{
			Status =  STATUS_DEVICE_CONFIGURATION_ERROR;
			break;
		}

	 	// record the base address
		pDevExt->MemPhysAddress = RegPhyAddressBase;
		pDevExt->MemPhysRange = RegPhyAddressLength;


		/************************************************************************/
		/* Do other initialization
		/************************************************************************/

		// Map the IO address
		pDevExt->MemVirAddress = MmMapIoSpace (
					pDevExt->MemPhysAddress,
					pDevExt->MemPhysRange,
					MmNonCached
					);
		
		if ( pDevExt->MemVirAddress == 0)
		{
			DBGPRINT(MP_INFO, ("Failed to allocate memory, not enough memory"));
			Status = STATUS_UNSUCCESSFUL;
			return Status;
		}

		// make sure the device is OK
		if ( ReadRegisterUlong (pDevExt, CPCI_BOARD_ID_REG) == 0xffffffff) 
		{
			Status = STATUS_UNSUCCESSFUL;
			return Status;
		}

		// set up DMA, the device supports packet-based, scatter/gather DMA operations, using 32-bit addressing
		WdfDeviceSetAlignmentRequirement ( pDevExt->Device, 32);

		WDF_DMA_ENABLER_CONFIG_INIT ( 
			&EnablerConfig,
			WdfDmaProfileScatterGather,
			MAX_DMA_LEN
			);

		Status = WdfDmaEnablerCreate ( 
					pDevExt->Device, &EnablerConfig,
					WDF_NO_OBJECT_ATTRIBUTES,
					&pDevExt->DMAEnabler
					);
		
		if ( Status != STATUS_SUCCESS) 
		{
			DBGPRINT(MP_INFO, ("Failed to set up the DMA advice"));
			Status = STATUS_UNSUCCESSFUL;
			return Status;
		}

		//
	   	// Allocate shared memory for send
	   	//
	   	Status = WdfCommonBufferCreate( 
	   				pDevExt->DMAEnabler,
					MAX_DMA_LEN,
					WDF_NO_OBJECT_ATTRIBUTES,
					&pDevExt->WdfSendCommonBuffer
					);

	  	if (Status != STATUS_SUCCESS)
	   	{
			DBGPRINT ( MP_INFO, ("WdfCommonBufferCreate(Send) failed %08X\n", Status ));
			break;
	   	}

	   	pDevExt->HwSendMemAllocVa = WdfCommonBufferGetAlignedVirtualAddress(
											   pDevExt->WdfSendCommonBuffer
											   );

	  	pDevExt->HwSendMemAllocPa = WdfCommonBufferGetAlignedLogicalAddress(
											   pDevExt->WdfSendCommonBuffer
											   );

	   	RtlZeroMemory(
			pDevExt->HwSendMemAllocVa,
			MAX_DMA_LEN
			);

		pDevExt->HwRcvMemAllocVa = MmAllocateContiguousMemorySpecifyCache( 
									MAX_DMA_LEN, 
									PhyLow, 
									PhyHigh, 
									PhyLow, 
									MmNonCached
									);
		
		if (pDevExt->HwRcvMemAllocVa == NULL)
	  	{
			DBGPRINT ( MP_INFO, ("MmAllocateContiguousMemorySpecifyCache(Receive) failed %08X\n", Status ));
			break;
	   	}

		pDevExt->HwRcvMemAllocPa = MmGetPhysicalAddress ( pDevExt->HwRcvMemAllocVa);

		RtlZeroMemory(
			pDevExt->HwRcvMemAllocVa,
			MAX_DMA_LEN
			);

	} while (FALSE);

	pDevExt->Iface = 0;

	//
	// Read additional info from NetFPGA such as MAC address
	//
	ReadAdapterInfo(pDevExt);

	DBGPRINT(MP_TRACE, ("<--- PCIEEvtDevicePrepareHardware"));

	return Status;
}

NTSTATUS
PCIEEvtDeviceReleaseHardware(
	IN	WDFDEVICE		Device,
	IN	WDFCMRESLIST	ResourcesTranslated
	)
/*++

Routine Description:

Unmap the resources that were mapped in PCIEEvtDevicePrepareHardware.
This will only be called when the Device stopped for resource rebalance,
surprise-removed or query-removed.

Arguments:

Device - A handle to the WDFDEVICE

ResourcesTranslated - The translated PnP resources associated with the
Device.  This is what is important to a PCI Device.

Return Value:

NT Status code - failure will result in the Device stack being torn down

--*/
{
	PDEVICE_EXTENSION	pDevExt = NULL;
	NTSTATUS			Status	= STATUS_SUCCESS;

	UNREFERENCED_PARAMETER(ResourcesTranslated);

	DBGPRINT(MP_TRACE, ("---> PCIEEvtDeviceReleaseHardware"));

	PAGED_CODE();

	pDevExt = PCIEGetDeviceContext(Device);

	do 
	{
		MmUnmapIoSpace (pDevExt->MemVirAddress, pDevExt->MemPhysRange);
	} while (FALSE);

	DBGPRINT(MP_TRACE, ("<--- PCIEEvtDeviceReleaseHardware"));

	return Status;
}
/*++

Routine Description:

This routine prepares the Device for use.  It is called whenever the Device
enters the D0 state, which happens when the Device is started, when it is
restarted, and when it has been powered off.

Note that interrupts will not be enabled at the time that this is called.
They will be enabled after this callback completes.

This function is not marked pageable because this function is in the
Device power up path. When a function is marked pagable and the code
section is paged out, it will generate a page fault which could impact
the fast resume behavior because the client driver will have to wait
until the system drivers can service this page fault.

Arguments:

Device	- The handle to the WDF Device object

PreviousState - The state the Device was in before this callback was invoked.

Return Value:

NTSTATUS

Success implies that the Device can be used.

Failure will result in the	  Device stack being torn down.

--*/
NTSTATUS
PCIEEvtDeviceD0Entry(
	IN	WDFDEVICE Device,
	IN	WDF_POWER_DEVICE_STATE PreviousState
	)
{

	PDEVICE_EXTENSION	pDevExt;
	NTSTATUS			Status;
	ULONG				mac_reset;
	ULONG				enable;

	DBGPRINT(MP_TRACE, ("---> PCIEEvtDeviceD0Entry"));

	UNREFERENCED_PARAMETER(PreviousState);

	pDevExt = PCIEGetDeviceContext(Device);

	NF2ResetAll(pDevExt);

	mac_reset = CNET_RESET_MAC;
	WriteRegisterUlong (pDevExt, CNET_REG_RESET, mac_reset);

	enable = ReadRegisterUlong (pDevExt, CNET_REG_ENABLE);
	DBGPRINT(MP_INFO, ("Enable Register before is x%x", ReadRegisterUlong (pDevExt, CNET_REG_ENABLE)));
	enable |=CNET_ENABLE_RX_FIFO | CNET_ENABLE_TX_MAC | 
		CNET_ENABLE_INGRESS_ARBITER | 
		CNET_ENABLE_RX_DMA;
	WriteRegisterUlong (pDevExt, CNET_REG_ENABLE, enable);

	DBGPRINT(MP_INFO, ("Enable Register now is x%x", ReadRegisterUlong (pDevExt, CNET_REG_ENABLE)));

	DBGPRINT(MP_TRACE, ("<--- PCIEEvtDeviceD0Entry"));

	return STATUS_SUCCESS;
}


/*++

Routine Description:

This routine undoes anything done in PCIEEvtDeviceD0Entry.	It is called
whenever the Device leaves the D0 state, which happens when the Device
is stopped, when it is removed, and when it is powered off.

The Device is still in D0 when this callback is invoked, which means that
the driver can still touch hardware in this routine.

Note that interrupts have already been disabled by the time that this
callback is invoked.

Arguments:

Device	- The handle to the WDF Device object

TargetState - The state the Device will go to when this callback completes.

Return Value:

Success implies that the Device can be used.  Failure will result in the
Device stack being torn down.

--*/
NTSTATUS
PCIEEvtDeviceD0Exit(
	IN	WDFDEVICE Device,
	IN	WDF_POWER_DEVICE_STATE TargetState
	)
{
	PDEVICE_EXTENSION	pDevExt;
	ULONG				mac_reset;
	ULONG				enable;

	DBGPRINT(MP_TRACE, ("---> PCIEEvtDeviceD0Exit"));

	PAGED_CODE();

	pDevExt = PCIEGetDeviceContext(Device);

	enable = ReadRegisterUlong (pDevExt, CNET_REG_ENABLE);
	DBGPRINT(MP_INFO, ("Enable Register before is x%x", ReadRegisterUlong (pDevExt, CNET_REG_ENABLE)));
	enable &=~(CNET_ENABLE_RX_FIFO | CNET_ENABLE_TX_MAC);
	WriteRegisterUlong (pDevExt, CNET_REG_ENABLE, enable);
	DBGPRINT(MP_INFO, ("Enable Register now is x%x", ReadRegisterUlong (pDevExt, CNET_REG_ENABLE)));

	mac_reset = CNET_RESET_MAC;
	WriteRegisterUlong (pDevExt, CNET_REG_RESET, mac_reset);

	NF2ResetAll(pDevExt);

	DBGPRINT(MP_TRACE, ("<--- PCIEEvtDeviceD0Exit"));

	return STATUS_SUCCESS;
}

/*++

Routine Description:

Reset the device to put the device in a known initial state.
This is called from D0Exit when the device is torn down or
when the system is shutdown. Note that Wdf has already
called out EvtDisable callback to disable the interrupt.

Arguments:

pDevExt -  Pointer to our adapter

Return Value:

None

--*/
VOID
PCIEDeviceShutdown(IN PDEVICE_EXTENSION pDevExt)
{
	//
	// WdfInterrupt is already disabled so issue a full reset
	//
	DBGPRINT(MP_TRACE, ("---> Device Shutdown <---"));
	
	PCIEDeviceReset(pDevExt);
	
}

VOID
PCIEDeviceReset(IN PDEVICE_EXTENSION pDevExt)
/*++
Routine Description:

Called by D0Exit when the device is being disabled or when the system is shutdown to
put the device in a known initial state.

Arguments:

DevExt	   Pointer to Device Extension

Return Value:

--*/
{
	UNREFERENCED_PARAMETER(pDevExt);

	DBGPRINT(MP_TRACE, ("---> PCIEDeviceReset <---"));
 
	return;
}


VOID PCIEDeviceWait(INT ms)
{
	LARGE_INTEGER delay;

	DBGPRINT(MP_TRACE, ("---> PCIEDeviceWait <---"));

	delay.QuadPart =  WDF_REL_TIMEOUT_IN_MS(ms);

	KeDelayExecutionThread( KernelMode, TRUE, &delay );
}



BOOLEAN
PciDrvReadRegistryKeyValue(
	IN	PWDFDEVICE_INIT  DeviceInit,
	IN	PWCHAR			 Name,
	OUT PULONG			 Value
	)
/*++

Routine Description:

	Can be used to read any REG_DWORD registry value stored
	under Device Parameter.

Arguments:

	FdoData - pointer to the device extension
	Name - Name of the registry value
	Value -


Return Value:

   TRUE if successful
   FALSE if not present/error in reading registry

--*/
{
	WDFKEY			hKey = NULL;
	NTSTATUS		status;
	BOOLEAN			retValue = FALSE;
	UNICODE_STRING	valueName;

	PAGED_CODE();

	DBGPRINT(MP_INFO, ("--->PciDrvReadRegistryKeyValue\n"));

	*Value = 0;

	status = WdfFdoInitOpenRegistryKey(DeviceInit,
								  PLUGPLAY_REGKEY_DEVICE,
								  STANDARD_RIGHTS_ALL,
								  WDF_NO_OBJECT_ATTRIBUTES,
								  &hKey
								  );

	if (NT_SUCCESS (status)) 
	{

		RtlInitUnicodeString(&valueName,Name);

		status = WdfRegistryQueryULong (hKey,
								  &valueName,
								  Value
								  );

		if (NT_SUCCESS (status)) 
		{
			retValue = TRUE;
		}

		WdfRegistryClose(hKey);
	}

	DBGPRINT(MP_INFO, ("<---PciDrvReadRegistryKeyValue\n"));

	return retValue;
}

void
ReadAdapterInfo( PDEVICE_EXTENSION pDevExt) 
{

	NTSTATUS	status;
	ULONG		val;
	PUSHORT		pShort;
	PULONG		pLong;
	int			i=0;

	// read permanent address
	val = ReadRegisterUlong (pDevExt, ROUTER_OP_LUT_MAC_0_HI_REG);
	pShort = (PUSHORT)(& pDevExt->PermanentAddress[4]);
	*pShort = (USHORT) val;

	val = ReadRegisterUlong (pDevExt, ROUTER_OP_LUT_MAC_0_LO_REG);
	pLong = (PULONG) pDevExt->PermanentAddress;
	*pLong = val;

	// current address should be the same as permanent address
	for ( i=0; i<ETH_LENGTH_OF_ADDRESS; i++) 
	{
		pDevExt->CurrentAddress[i] = pDevExt->PermanentAddress[ETH_LENGTH_OF_ADDRESS-i-1];
	}

	for ( i=0; i<ETH_LENGTH_OF_ADDRESS; i++) 
	{
		pDevExt->PermanentAddress[i] = pDevExt->CurrentAddress[i];
	}

	DBGPRINT(MP_INFO, ("Current Address = %02x-%02x-%02x-%02x-%02x-%02x\n",
									pDevExt->CurrentAddress[0],
									pDevExt->CurrentAddress[1],
									pDevExt->CurrentAddress[2],
									pDevExt->CurrentAddress[3],
									pDevExt->CurrentAddress[4],
									pDevExt->CurrentAddress[5]));

}

ULONG
ReadRegisterUlong (IN PDEVICE_EXTENSION pDevExt, IN ULONG address)
{
	return READ_REGISTER_ULONG ( (PULONG)((ULONG_PTR)pDevExt->MemVirAddress + address) );
}

VOID
WriteRegisterUlong (IN PDEVICE_EXTENSION pDevExt, IN ULONG address, IN ULONG value) 
{
	WRITE_REGISTER_ULONG ( (PULONG)((ULONG_PTR)pDevExt->MemVirAddress + address), value);
}