/*++

Copyright (c) 2009	Microsoft Corporation

Module Name:
	IsrDpc.c

Abstract:
	Interrupt Handler

Revision History:
	Who 		When		What						
	--------	--------	-----------------------------------------------
	Chen Tian	16-03-09	Wrtitten For NetFPGA2 Card in Microsoft Research Asia
Notes:

--*/

#include "..\inc\precomp.h"
#include "..\inc\Device.h"
#include "..\inc\nf2.h"
#include "..\inc\IsrDpc.h"
#include "..\inc\dbg.h"
#include "..\inc\reg_defines.h"

/*++
Routine Description:

Configure and create the WDFINTERRUPT object.
This routine is called by EvtDeviceAdd callback.

Arguments:

pDevExt 	 Pointer to our DEVICE_EXTENSION

Return Value:

NTSTATUS code

--*/
NTSTATUS
PCIEInterruptCreate(IN PDEVICE_EXTENSION pDevExt)
{
	NTSTATUS					Status;
	WDF_INTERRUPT_CONFIG		InterruptConfig;

	DBGPRINT(MP_TRACE, ("---> PCIEInterruptCreate\n"));

	WDF_INTERRUPT_CONFIG_INIT( &InterruptConfig,
							   PCIEEvtInterruptIsr,
							   PCIEEvtInterruptDpc 
							   );

	InterruptConfig.EvtInterruptEnable	= NF2InterruptEnable;
	InterruptConfig.EvtInterruptDisable 	= NF2InterruptDisable;

	InterruptConfig.AutomaticSerialization = TRUE;

	//
	// Unlike WDM, framework driver should create interrupt object in EvtDeviceAdd and
	// let the framework do the resource parsing and registration of ISR with the kernel.
	// Framework connects the interrupt after invoking the EvtDeviceD0Entry callback
	// and disconnect before invoking EvtDeviceD0Exit. EvtInterruptEnable is called after
	// the interrupt interrupt is connected and EvtInterruptDisable before the interrupt is
	// disconnected.
	//
	Status = WdfInterruptCreate( 
				pDevExt->Device, 
				&InterruptConfig, 
				WDF_NO_OBJECT_ATTRIBUTES, 
				&pDevExt->Interrupt );

	if( !NT_SUCCESS(Status) )
	{
		KdPrint(("PCIEInterruptCreate: Failed: %x", Status));
	}
	
	DBGPRINT(MP_INFO, ("<--- PCIEInterruptCreate Status %d\n",Status));
	
	return Status;
}


/*++
Routine Description:

Interrupt handler for this driver. Called at DIRQL level when the
device or another device sharing the same interrupt line asserts
the interrupt. The driver first checks the device to make sure whether
this interrupt is generated by its device and if so clear the interrupt
register to disable further generation of interrupts and queue a
DPC to do other I/O work related to interrupt - such as reading
the device memory, starting a DMA transaction, coping it to
the request buffer and completing the request, etc.

Arguments:

Interupt   - Handle to WDFINTERRUPT Object for this device.
MessageID  - MSI message ID (always 0 in this configuration)

Return Value:

TRUE   --  This device generated the interrupt.
FALSE  --  This device did not generated this interrupt.

--*/
BOOLEAN
PCIEEvtInterruptIsr(IN WDFINTERRUPT Interrupt, IN ULONG MessageID)
{
	PDEVICE_EXTENSION	pDevExt=NULL;
	BOOLEAN				IsRecognized = FALSE;
	ULONG				mask;
	ULONG				status;
	NTSTATUS			ret;
	ULONG				err;

	UNREFERENCED_PARAMETER(MessageID);

	// Our ISR Routine will be called event if the test card hasn't trigger any interrupt signal.
	pDevExt  = PCIEGetDeviceContext(WdfInterruptGetDevice(Interrupt));

	/* get the interrupt mask */
	mask = READ_REGISTER_ULONG ( (PULONG)((ULONG_PTR)pDevExt->MemVirAddress + CPCI_REG_INTERRUPT_MASK) );

	//
	// Disable the interrupt (will be re-enabled in NICEvtInterruptDpc
	//
	WRITE_REGISTER_ULONG ( (PULONG)((ULONG_PTR)pDevExt->MemVirAddress + CPCI_REG_INTERRUPT_MASK), 0xFFFFFFFF);


	/* Grab the interrupt status */
	status = READ_REGISTER_ULONG ( (PULONG)((ULONG_PTR)pDevExt->MemVirAddress + CPCI_REG_INTERRUPT_STATUS) );

	DBGPRINT(MP_TRACE, ("---> PCIEEvtInterruptIsr <---. mask %x, status %x", mask, status));

	/* only consider bits that are not masked plus INT_PKT_AVAIL*/
	status &= ~(mask & ~INT_PKT_AVAIL); 
	 
	if (status && (mask!=0xffffffff))
	{
		pDevExt->Status = status;
		pDevExt->Mask = mask;
		WdfInterruptQueueDpcForIsr( Interrupt );
		IsRecognized = TRUE;
	}
	else
	{
		WRITE_REGISTER_ULONG ( (PULONG)((ULONG_PTR)pDevExt->MemVirAddress + CPCI_REG_INTERRUPT_MASK), mask);
	}

	return IsRecognized;
}

VOID
PCIEEvtInterruptDpc(IN WDFINTERRUPT Interrupt,IN WDFDEVICE Device)
/*++

Routine Description:

	DPC callback for ISR. Please note that on a multiprocessor system,
	you could have more than one DPCs running simulataneously on
	multiple processors. So if you are accesing any global resources
	make sure to synchrnonize the accesses with a spinlock.

Arguments:

	Interupt  - Handle to WDFINTERRUPT Object for this device.
	Device	  - WDFDEVICE object passed to InterruptCreate

Return Value:

--*/
{
	PDEVICE_EXTENSION	pDevExt=NULL;
	ULONG				mask;
	ULONG				status;
	NTSTATUS			ret;
	ULONG				err;
	BOOLEAN				transactionComplete= FALSE;
	WDFDMATRANSACTION	dmaTransaction;
	WDFREQUEST			Request;
	size_t				length;
	PMDL				mdl;
	PVOID				RecvBufferVa;
	NDIS_PHYSICAL_ADDRESS	RecvBufferPa;
	ULONG				RecvBufferSize;
	ULONG				InterfaceNo=0;

	// Our ISR Routine will be called event if the test card hasn't trigger any interrupt signal.
	pDevExt  = PCIEGetDeviceContext(WdfInterruptGetDevice(Interrupt));

	status = pDevExt->Status;

	mask = pDevExt->Mask;

	DBGPRINT(MP_TRACE, ("---> PCIEEvtInterruptDpc mask %x, status %x", mask, status));

	/* Check if the interrrupt was generated by us */
	if (status && (mask!=0xffffffff))
	{
		   
	  	/* Handle packet RX complete 
	   	* Note: don't care about the rx pool here as the packet is copied to an skb
	   	* immediately so there is no need to have multiple packets in the rx pool
	   	*/
	   	if (status & INT_DMA_RX_COMPLETE)
		{
			DBGPRINT(MP_LOUD, ("nf2: intr: INT_DMA_RX_COMPLETE\n"));

			// since the data is in the continuous buffer we allocated now, we just need to copy it to the receive buffer
			WdfInterruptAcquireLock( Interrupt );
			length = READ_REGISTER_ULONG ( (PULONG)((ULONG_PTR)pDevExt->MemVirAddress + CPCI_REG_DMA_I_SIZE) );
			InterfaceNo = READ_REGISTER_ULONG ( (PULONG)((ULONG_PTR)pDevExt->MemVirAddress + CPCI_REG_DMA_I_CTRL) );
			WdfInterruptReleaseLock( Interrupt );
			// get the interface no
		 	InterfaceNo = (InterfaceNo & 0x300) >> 8;
		 	DBGPRINT(MP_LOUD, (" Received interface is %x,length is %x", InterfaceNo,length));
		 			
			length = (InterfaceNo<<30) | (length&0x3fffffff);
			DBGPRINT(MP_LOUD, ("now length is %x", length));
			WdfRequestCompleteWithInformation( pDevExt->currentRequest, STATUS_SUCCESS, length);
			pDevExt->currentRequest = NULL;
			 	
		 	InterlockedDecrement (&pDevExt->RecvinProgress);

			mask &= ~INT_PKT_AVAIL;
	  	}
		   
	  	/* Handle packet TX complete */
		if (status & INT_DMA_TX_COMPLETE ) 
		{
			DBGPRINT(MP_LOUD, ("nf2: intr: INT_DMA_TX_COMPLETE\n"));

			if (pDevExt->SendinProgress) 
			{
		  		//
		  		// Indicate this DMA operation has completed:
		  		// This may drive the transfer on the next packet if
		  		// there is still data to be transfered in the DmaTransaction.
		  		//
		  		dmaTransaction = pDevExt->currentDmaTransaction;
		  		pDevExt->currentDmaTransaction = NULL;

		  		if (dmaTransaction)
					transactionComplete = WdfDmaTransactionDmaCompleted( dmaTransaction,	&ret );

		  		if(transactionComplete == TRUE) 
				{
			  		ASSERT(ret == STATUS_SUCCESS);
			  		Request = WdfDmaTransactionGetRequest(dmaTransaction);
			  		length  = WdfDmaTransactionGetBytesTransferred(dmaTransaction);

			  		WdfObjectDelete( dmaTransaction );

			  		if (Request)
			  		{
			  			WdfRequestCompleteWithInformation(Request, ret, length);
			  		}
			  		else 
			  		{
				  		DBGPRINT(MP_INFO, ("***************** Ooooops, it failed"));
			  		}
			 	} 
		  		else 
				{
				 	//
				 	// NOTE: For this ethernet driver this should never
				 	//	   be returned as the packets are <= 1514 bytes.
				 	//	   It is included to show the complete DmaTransaction
				 	//	   coding pattern.
				 	//
				 	//ASSERT(!"STATUS_MORE_PROCESSING_REQUIRED");
				}
			 	InterlockedDecrement(&pDevExt->SendinProgress);
		  	}
	  	}

		/* Handle PHY interrupts */
		if (status & INT_PHY_INTERRUPT)
	  	{
		  	DBGPRINT(MP_INFO, ("nf2: intr: INT_PHY_INTERRUPT"));
		  	mask |= INT_PHY_INTERRUPT;
		  	DBGPRINT(MP_INFO, ("nf2: Phy Interrrupt - masked off"));
	  	}

	 	/* Handle a packet RX notification 
	 	*
	 	* Should not need to worry about this interrupt being asserted while
	 	* a DMA transfer is in progress as the hardware should prevent this.
	 	*
	 	* ie. no need to do: !card->dma_rx_in_progress
	 	*/
		if (status & INT_PKT_AVAIL) 
		{
			DBGPRINT(MP_LOUD, ("nf2: intr: INT_PKT_AVAIL\n"));

			if ( InterlockedIncrement(&pDevExt->RecvinProgress)==1) 
			{
				ret = WdfIoQueueRetrieveNextRequest( pDevExt->PendingReadQueue, &Request );
			
				//
				// Initialize this new DmaTransaction with direct parameters.
				//
				if(NT_SUCCESS(ret))
				{
					ret = WdfRequestRetrieveOutputWdmMdl(Request, &mdl);
					if (!NT_SUCCESS(ret)) 
					{
						DBGPRINT(MP_INFO,
							("WdfRequestRetrieveInputWdmMdl failed: %x", ret));
						WdfRequestComplete(Request, STATUS_UNSUCCESSFUL);
						InterlockedDecrement(&pDevExt->RecvinProgress);
					}
					else
					{
						pDevExt->currentRequest = Request;

						RecvBufferVa = MmGetMdlVirtualAddress(mdl);
						RecvBufferPa = MmGetPhysicalAddress (RecvBufferVa);

						/* Start the transfer */
						WdfInterruptAcquireLock( Interrupt );
						WRITE_REGISTER_ULONG ( (PULONG)((ULONG_PTR)pDevExt->MemVirAddress + CPCI_DMA_ADDR_I_REG), RecvBufferPa.LowPart);
						WRITE_REGISTER_ULONG ( (PULONG)((ULONG_PTR)pDevExt->MemVirAddress + CPCI_DMA_CTRL_I_REG), DMA_CTRL_OWNER);
						WdfInterruptReleaseLock( Interrupt );
					}
				}
				else 
				{
				 	DBGPRINT(MP_WARN, ("No request %x!", ret));
				 	InterlockedDecrement(&pDevExt->RecvinProgress);
				}
			}
			else 
			{	
				DBGPRINT(MP_INFO, ("nf2: received interrupt for new rx packet avail while still %x,\n",pDevExt->RecvinProgress));
				InterlockedDecrement(&pDevExt->RecvinProgress);
			}

			/* mask off subsequent PKT_AVAIL interrupts */
			mask |= INT_PKT_AVAIL;
		}

	  	/* DMA transfer error */
	 	if (status & (INT_DMA_TRANSFER_ERROR|INT_DMA_SETUP_ERROR|INT_DMA_FATAL_ERROR))
	  	{
		  	err = ReadRegisterUlong (pDevExt, CPCI_REG_ERROR);
			
		  	DBGPRINT(MP_WARN, ("nf2: DMA transfer error: 0x%08x\n", err));
			
		  	NF2ResetCPCI(pDevExt);

		 	if (pDevExt->SendinProgress) 
			{
				dmaTransaction = pDevExt->currentDmaTransaction;
				pDevExt->currentDmaTransaction = NULL;

				if ( dmaTransaction ) 
				{
			  		// error, release this request
			  		Request = WdfDmaTransactionGetRequest(dmaTransaction);

			  		WdfDmaTransactionDmaCompletedFinal(dmaTransaction, 0, &ret);

			  		WdfObjectDelete(dmaTransaction);        

			  		if (Request)
						WdfRequestCompleteWithInformation( Request, STATUS_UNSUCCESSFUL, 0);
				}
				InterlockedDecrement(&pDevExt->SendinProgress);
			}
	 	}

	  	/* Rewrite the interrupt mask including any changes */
		WdfInterruptAcquireLock( Interrupt );
	  	WRITE_REGISTER_ULONG ( (PULONG)((ULONG_PTR)pDevExt->MemVirAddress + CPCI_REG_INTERRUPT_MASK), mask);
		WdfInterruptReleaseLock( Interrupt );
	  
	}

	DBGPRINT(MP_TRACE, ("<--- PCIEEvtInterruptDpc mask %x", mask));
	return;
}


