/*++

Copyright (c) 2009	Microsoft Corporation

Module Name:
	Write.c

Abstract:
	PCI write operations

Revision History:
	Who 		When		What						
	--------	--------	-----------------------------------------------
	Chen Tian	16-03-09	Wrtitten For NetFPGA2 Card in Microsoft Research Asia
Notes:

--*/
#include "..\inc\precomp.h"
#include "..\inc\dbg.h"
#include "..\inc\device.h"
#include "..\inc\nf2.h"
#include "..\inc\reg_defines.h"

ULONG		InterfaceNo = 0;

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
VOID
PciDrvEvtIoWrite(
	IN WDFQUEUE 		Queue,
	IN WDFREQUEST		Request,
	IN size_t			Length
	)
/*++

Routine Description:

	Called by the framework as soon as it receives a write request.
	If the device is not ready, fail the request.
	Otherwise get scatter-gather list for this request and send the
	packet to the hardware for DMA.

Arguments:

	Queue - Handle to the framework queue object that is associated
			with the I/O request.
	Request - Handle to a framework request object.

	Length - Length of the IO operation
				 The default property of the queue is to not dispatch
				 zero lenght read & write requests to the driver and
				 complete is with status success. So we will never get
				 a zero length request.

Return Value:

--*/
{
	NTSTATUS			status = STATUS_SUCCESS;
	PDEVICE_EXTENSION	pDevExt=NULL;
	WDFDMATRANSACTION	writeDmaTransaction;
	PRegDes 				pReg;
	PMDL				mdl;
	PVOID				virtualAddress;
	PNDIS_PACKET		Packet = NULL;

	DBGPRINT(MP_INFO, ("---> PciDrvEvtIoWrite"));

	//
	// Get the pDevExt from the Queue handle
	//
	pDevExt = PCIEGetDeviceContext(WdfIoQueueGetDevice(Queue));

	//
	// Following code illustrates two different ways of initializing a DMA
	// transaction object. If ASSOC_WRITE_REQUEST_WITH_DMA_TRANSACTION is
	// defined in the sources file, the first section will be used.
	//
#ifdef ASSOC_WRITE_REQUEST_WITH_DMA_TRANSACTION

	//
	// Initialize this new DmaTransaction with direct parameters.
	//
	
	InterfaceNo  = (ULONG)WdfRequestGetInformation (Request);;
	DBGPRINT(MP_INFO, ("InterfaceNo is 0x%x", InterfaceNo));

	//
	// This section illustrates how to create and initialize
	// a DmaTransaction using a WDF Request.
	// This type of coding pattern would probably be the most commonly used
	// for handling client Requests.
	//
	status = WdfDmaTransactionCreate( 
				pDevExt->DMAEnabler, 
				WDF_NO_OBJECT_ATTRIBUTES, 
				&writeDmaTransaction
				);
	
	if(!NT_SUCCESS(status)) 
	{
		return;
	}

	status = WdfDmaTransactionInitializeUsingRequest(
						writeDmaTransaction,
						Request,
						PCIEEvtProgramWriteDma,
						WdfDmaDirectionWriteToDevice 
						);

	if(!NT_SUCCESS(status)) 
	{
		DBGPRINT(MP_ERROR, (
					"WdfDmaTransactionInitializeUsingRequest failed: "
					"%!STATUS!", status));
		goto CleanUp;
	}
#else
	//
	// This section illustrates how to create and initialize
	// a DmaTransaction via direct parameters (e.g. not using a WDF Request).
	// This type of coding pattern might be used for driver-initiated DMA
	// operations (e.g. DMA operations not based on driver client requests.)
	//
	// NOTE: This example unpacks the WDF Request in order to get a set of
	//		 parameters for the call to WdfDmaTransactionInitialize. While
	//		 this is completely legimate, the represenative usage pattern
	//		 for WdfDmaTransactionIniitalize would have the driver create/
	//		 initialized a DmaTransactin without a WDF Request. A simple
	//		 example might be where the driver needs to DMA the devices's
	//		 firmware to it during device initialization. There would be
	//		 no WDF Request; the driver would supply the parameters for
	//		 WdfDmaTransactionInitialize directly.
	//
	{
		PMDL				  mdl;
		PVOID				  virtualAddress;
		ULONG				  length;

		//
		// Initialize this new DmaTransaction with direct parameters.
		//
		status = WdfRequestRetrieveInputWdmMdl(Request, &mdl);
		if (!NT_SUCCESS(status)) {
			DBGPRINT(MP_INFO,
						("WdfRequestRetrieveInputWdmMdl failed: %!STATUS!", status));
			goto CleanUp;
		}

		virtualAddress = MmGetMdlVirtualAddress(mdl);
		length = MmGetMdlByteCount(mdl);

		status = WdfDmaTransactionCreate( pDevExt->DMAEnabler, 
			WDF_NO_OBJECT_ATTRIBUTES, 
			&writeDmaTransaction);
		if(!NT_SUCCESS(status)) {
			return;
		}


		status = WdfDmaTransactionInitialize( writeDmaTransaction,
											  PCIEEvtProgramWriteDma,
											  WdfDmaDirectionWriteToDevice,
											  mdl,
											  virtualAddress,
											  length );

		if(!NT_SUCCESS(status)) {
			DBGPRINT(MP_INFO,
						("WdfDmaTransactionInitialize failed: %!STATUS!", status));
			  goto CleanUp;
		}

	}
#endif

	pDevExt->currentDmaTransaction = writeDmaTransaction;
	//
	// Execute this DmaTransaction transaction.
	//
	status = WdfDmaTransactionExecute( 
				writeDmaTransaction, 
				WDF_NO_CONTEXT
				);

	if(!NT_SUCCESS(status)) 
	{
		//
		// Couldn't execute this DmaTransaction, so fail Request.
		//
	   	DBGPRINT(MP_INFO,
					("WdfDmaTransactionExecute failed: %!STATUS!", status));
		goto CleanUp;
	}

	//
	// Indicate that Dma transaction has been started successfully. The request
	// will be complete by the Dpc routine when the DMA transaction completes.
	//
	status = STATUS_SUCCESS;

CleanUp:

	//
	// If there are errors, then clean up and complete the Request.
	//
	if (!NT_SUCCESS(status)) 
	{
		WdfDmaTransactionRelease(writeDmaTransaction);		  
		WdfRequestComplete(Request, status);
	}
	
	DBGPRINT(MP_INFO, ("<--- PciDrvEvtIoWrite"));

}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
BOOLEAN
PCIEEvtProgramWriteDma(
	IN	WDFDMATRANSACTION		Transaction,
	IN	WDFDEVICE				Device,
	IN	PVOID					Context,
	IN	WDF_DMA_DIRECTION		Direction,
	IN	PSCATTER_GATHER_LIST	SgList
	)
/*++

Routine Description:

Arguments:

Return Value:

--*/
{
	PDEVICE_EXTENSION		 pDevExt;
	size_t					 offset;
	ULONG					 AddressLow;
	BOOLEAN 				 errors;
	ULONG					 length;
	ULONG					 i;
	ULONG					 status;
	WDFREQUEST				 Request;

	UNREFERENCED_PARAMETER( Context );
	UNREFERENCED_PARAMETER( Direction );

	DBGPRINT(MP_INFO, ("---> PCIEEvtProgramWriteDma"));

	//
	// Initialize locals
	//
	pDevExt = PCIEGetDeviceContext(Device);
	errors = FALSE;

	//
	// Start the DMA operation.
	if ( InterlockedIncrement(&pDevExt->SendinProgress)!=1) 
	{
		InterlockedDecrement(&pDevExt->SendinProgress);

		Request = WdfDmaTransactionGetRequest(Transaction);

		(VOID) WdfDmaTransactionDmaCompletedFinal(Transaction, 0, &status);
		ASSERT(NT_SUCCESS(status));

		WdfDmaTransactionRelease(Transaction);		  

		WdfRequestCompleteWithInformation( Request, STATUS_UNSUCCESSFUL, 0);

		DBGPRINT(MP_WARN, ("<-- PCIEEvtProgramWriteDma: error ****"));
		return FALSE;

	}

	//
	// Get the number of bytes as the offset to the beginning of this
	// Dma operations transfer location in the buffer.
	//
	offset = WdfDmaTransactionGetBytesTransferred(Transaction);

	//
	// Translate the System's SCATTER_GATHER_LIST elements
	// into the device's DMA_TRANSFER_ELEMENT elements.
	//

	for (i=0; i < SgList->NumberOfElements; i++) 
	{
		//
		// Construct this DTE.
		//
		// NOTE: The LocalAddress is the offset into the SRAM from
		//		 where this Write will start.
		//
		AddressLow			  = SgList->Elements[i].Address.LowPart;
		length				  = SgList->Elements[i].Length;

		//
		// Increment the DmaTransaction length by this element length
		//
		offset += SgList->Elements[i].Length;

		/* Start the transfer */
		
		WdfInterruptAcquireLock( pDevExt->Interrupt );
		WriteRegisterUlong (pDevExt, CPCI_DMA_ADDR_E_REG, AddressLow);
		WriteRegisterUlong (pDevExt, CPCI_DMA_SIZE_E_REG, length);
		WriteRegisterUlong (pDevExt, CPCI_DMA_CTRL_E_REG, NF2_SET_DMA_CTRL_MAC(InterfaceNo) | DMA_CTRL_OWNER);
		WdfInterruptReleaseLock( pDevExt->Interrupt );

	}

	DBGPRINT(MP_INFO, ("<--- PCIEEvtProgramWriteDma"));

	return TRUE;
}


VOID
PCIEWriteRequestComplete(
	IN WDFDMATRANSACTION  DmaTransaction,
	IN NTSTATUS 		  Status
	)
/*++

Routine Description:

Arguments:

Return Value:

--*/
{
	WDFDEVICE		   device;
	WDFREQUEST		   request;
	PDEVICE_EXTENSION  pDevExt;
	size_t			   bytesTransferred;

	//
	// Initialize locals
	//
	device	= WdfDmaTransactionGetDevice(DmaTransaction);
	pDevExt  = PCIEGetDeviceContext(device);

	request = WdfDmaTransactionGetRequest(DmaTransaction);

	//
	// Get the final bytes transferred count.
	//
	bytesTransferred =	WdfDmaTransactionGetBytesTransferred( DmaTransaction );

	DBGPRINT ( MP_INFO, (
				"PCIEWriteRequestComplete:	Request %p, Status %!STATUS!, "
				"bytes transferred %d\n",
				 request, Status, (int) bytesTransferred ));

	WdfDmaTransactionRelease(DmaTransaction);		 

	WdfRequestCompleteWithInformation( request, Status, bytesTransferred);

}


